unit Forces;

{$mode objfpc}{$H+}

interface

uses
  Physics;

type

  { TForce }

  TForce = class
  protected
    function GetValueFor(const body: TPhysicsBody): TVector3; virtual; abstract;
  public
    procedure Update(const delta: Double); virtual;
    property Value[body: TPhysicsBody]: TVector3 read GetValueFor; default;
  end;

  { TPlanetsGravity }

  TPlanetsGravity = class(TForce)
  private
    FGValue: Double;     
  protected
    function GetValueFor(const body: TPhysicsBody): TVector3; override;
  public
    constructor Create(const GValue: Double = 9.81);
    property G: Double read FGValue;
  end;

  { TNewtonGravity }

  TNewtonGravity = class(TForce)
  private
    FReferenceBodies: TPhysicsBodies;
    procedure SetReferenceBodies(AValue: TPhysicsBodies);
  protected
    function GetValueFor(const body: TPhysicsBody): TVector3; override;
  public
    constructor Create;
    destructor Destroy; override;
    property ReferenceBodies: TPhysicsBodies read FReferenceBodies write SetReferenceBodies;
  end;

  { TDrag }

  TDrag = class(TForce)
  private
    FRoh: Double; // density of fluid
  protected
    function GetValueFor(const body: TPhysicsBody): TVector3; override;
  public
    constructor Create(const Roh: Double);
    property Roh: Double read FRoh write FRoh;
  end;

implementation

{ TDrag }

function TDrag.GetValueFor(const body: TPhysicsBody): TVector3;
var
  DragCoefficient, v: Double;
  direction: TVector3;
  vBody: TVolumeBody;
begin
  // Drag coefficient depends on body type
  // see https://www.wikiwand.com/en/Drag_coefficient
  if body is TSphereBody then
    DragCoefficient := 0.47
  else
    DragCoefficient := 0; // unsupported body type
  vBody := body as TVolumeBody;
  v := body.Velocity.Length;
  direction := body.Velocity.Normalized;
  // See https://www.wikiwand.com/en/Drag_(physics)
  Result := -direction * 1/2 * FRoh * sqr(v) * DragCoefficient * vBody.GetSurfaceAreaFrom(direction);
end;

constructor TDrag.Create(const Roh: Double);
begin
  inherited Create;
  FRoh := Roh;
end;

{ TNewtonGravity }

procedure TNewtonGravity.SetReferenceBodies(AValue: TPhysicsBodies);
begin
  if FReferenceBodies=AValue then Exit;
  FReferenceBodies.Assign(AValue);
end;

function TNewtonGravity.GetValueFor(const body: TPhysicsBody): TVector3;
var
  ref: TPhysicsBody;
  distance: TVector3;
  force, r: Double;
const
  G: Double = 6.674e-11;
begin
  Result := TVector3.Zero;
  for ref in ReferenceBodies do
  begin
    distance := ref.Postion - body.Postion;
    r := distance.Length;
    if r = 0 then Continue;
    force := G * body.Mass * ref.Mass / sqr(r);
    Result += distance.Normalized * force;
  end;
end;

constructor TNewtonGravity.Create;
begin
  inherited Create;
  FReferenceBodies := TPhysicsBodies.Create;
end;

destructor TNewtonGravity.Destroy;
begin
  FReferenceBodies.Free;
  inherited Destroy;
end;

{ TPlanetsGravity }

constructor TPlanetsGravity.Create(const GValue: Double);
begin        
  inherited Create;
  FGValue:=GValue;
end;

function TPlanetsGravity.GetValueFor(const body: TPhysicsBody): TVector3;
begin
  Result := Vector3(0, -G * body.Mass, 0);
end;

{ TForce }

procedure TForce.Update(const delta: Double);
begin
  // Noop
end;

end.

