unit Physics;

{$mode objfpc}{$H+}
{$modeSwitch advancedRecords}
interface

uses
  fgl;

type

  { TVector3 }

  TVector3 = record
    X, Y, Z: double;
    function Length: double;
    function Normalized: TVector3;
    class function Zero: TVector3; static;
  end;

  { TPhysicsBody }

  TPhysicsBody = class
  private
    FPosition: TVector3;
    FVelocity: TVector3;
    FAcceleration: TVector3;
    FMass: Double;
  public
    constructor Create(const AMass: Double; const APosition: TVector3; const InitialVelocity: TVector3);

    procedure ResetForces; inline;
    procedure ApplyForce(Force: TVector3); inline;
    procedure Update(const Delta: Double); inline;

    property Postion: TVector3 read FPosition write FPosition;
    property Velocity: TVector3 read FVelocity write FVelocity;
    property Acceleration: TVector3 read FAcceleration write FAcceleration;
    property Mass: Double read FMass write FMass;
  end;

  TPhysicsBodies = class(specialize TFPGList<TPhysicsBody>);

  { TVolumeBody }

  TVolumeBody = class(TPhysicsBody)
  protected
    function GetVolume: Double; virtual; abstract;
    function GetSurfaceArea: Double; virtual; abstract;
    function GetDensity: Double;
  public
    // See https://www.wikiwand.com/en/Cross_section_(geometry)
    function GetSurfaceAreaFrom(Direction: TVector3): Double; virtual; abstract;

    property Volume: Double read GetVolume;
    property SurfaceArea: Double read GetSurfaceArea;
    property Density: Double read GetDensity;
  end;

  TVolumeBodies = class(specialize TFPGList<TVolumeBody>);

  { TSphereBody }

  TSphereBody = class(TVolumeBody)
  private
    FRadius: Double;
  protected
    function GetVolume: Double; override;
    function GetSurfaceArea: Double; override;
  public      
    constructor Create(const ARadius: Double; const AMass: Double;
      const APosition: TVector3; const InitialVelocity: TVector3);

    function GetSurfaceAreaFrom(Direction: TVector3): Double; override;

    property Radius: Double read FRadius write FRadius;
  end;

operator +(const lhs: TVector3; const rhs: TVector3): TVector3; inline;
operator -(const lhs: TVector3; const rhs: TVector3): TVector3; inline;
operator -(const rhs: TVector3): TVector3; inline;
operator * (const lhs: TVector3; const rhs: double): TVector3; inline;
operator / (const lhs: TVector3; const rhs: double): TVector3; inline;
function Vector3(const X: double; const Y: double; const Z: double): TVector3; inline;
function Vector2(const X: double; const Y: double): TVector3; inline;

implementation

operator +(const lhs: TVector3; const rhs: TVector3): TVector3;
begin
  Result := Vector3(lhs.X + rhs.X, lhs.Y + rhs.Y, lhs.z + rhs.z);
end;

operator -(const lhs: TVector3; const rhs: TVector3): TVector3;
begin
  Result := Vector3(lhs.X - rhs.X, lhs.Y - rhs.Y, lhs.z - rhs.z);
end;

operator-(const rhs: TVector3): TVector3;
begin
  Result := Vector3(-rhs.X, -rhs.Y, -rhs.Z);
end;

operator * (const lhs: TVector3; const rhs: double): TVector3;
begin
  Result := Vector3(lhs.X * rhs, lhs.Y * rhs, lhs.Z * rhs);
end;

operator / (const lhs: TVector3; const rhs: double): TVector3;
begin
  Result := Vector3(lhs.X / rhs, lhs.Y / rhs, lhs.Z / rhs);
end;

function Vector3(const X: double; const Y: double; const Z: double): TVector3;
begin
  Result.X := X;
  Result.Y := Y;
  Result.Z := Z;
end;

function Vector2(const X: double; const Y: double): TVector3;
begin
  Result := Vector3(X, Y, 0);
end;

{ TVolumeBody }

function TVolumeBody.GetDensity: Double;
begin
  Result := Mass / Volume;
end;

{ TPhysicsBody }

constructor TPhysicsBody.Create(const AMass: Double; const APosition: TVector3;
  const InitialVelocity: TVector3);
begin
  Self.Mass := AMass;
  Self.Postion := APosition;
  Self.Velocity := InitialVelocity;
  Self.ResetForces;
end;

procedure TPhysicsBody.ResetForces;
begin
  FAcceleration := TVector3.Zero;
end;

procedure TPhysicsBody.ApplyForce(Force: TVector3);
begin
  FAcceleration += Force / Mass;
end;

procedure TPhysicsBody.Update(const Delta: Double);
begin
  FVelocity += Acceleration * Delta;
  FPosition += Self.Velocity * Delta;
end;

{ TSphereBody }

function TSphereBody.GetVolume: Double;
begin
  Result := 3/4 * pi * Radius*Radius*Radius;
end;

function TSphereBody.GetSurfaceArea: Double;
begin
  Result := 4 * pi * Radius*Radius;
end;

constructor TSphereBody.Create(const ARadius: Double; const AMass: Double;
  const APosition: TVector3; const InitialVelocity: TVector3);
begin
  inherited Create(AMass, APosition, InitialVelocity);
  FRadius := ARadius;
end;

function TSphereBody.GetSurfaceAreaFrom(Direction: TVector3): Double;
begin
  // Projected Surface of a sphere is identical from all angles
  Result := pi * Radius * Radius;
end;

{ TVector3 }

function TVector3.Length: double;
begin
  Result := sqrt(sqr(Self.X) + sqr(Self.Y) + sqr(Self.Z));
end;

function TVector3.Normalized: TVector3;
var
  len: Double;
begin
  len := Self.Length;
  if len > 0 then
    Result := Self / len;
end;

class function TVector3.Zero: TVector3;
begin
  Result := Vector3(0, 0, 0);
end;

end.
